from aitemplates.oai.utils.wrappers import retry_openai_api
import os
from dotenv import load_dotenv

from typing import Any, Optional, List, Union

import aiolimiter
import openai
import openai.error
from aiohttp import ClientSession
from tqdm.asyncio import tqdm_asyncio

from aitemplates.oai.types.base import ResponseDict
from aitemplates.oai.types.chat import ChatConversation, FunctionsAvailable, ChatPair
from aitemplates.oai.ApiManager import SingleApiManager

dotenv_path = os.path.join(os.getcwd(), '.env')  # get the path to .env file in current working directory
load_dotenv(dotenv_path)  # load environment variables from the .env file

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

openai.api_key = OPENAI_API_KEY


@retry_openai_api()
async def _throttled_acreate_chat_completion(
    chat_pair: ChatPair,
    limiter: aiolimiter.AsyncLimiter,
    **kwargs
) -> Any:
    """Create a throttled chat completion using the OpenAI API.

    This function ensures the number of requests made to the OpenAI API does not exceed a specified limit.

    Args:
        messages (list[MessageDict]): The messages to send to the chat completion.
        limiter (aiolimiter.AsyncLimiter): The limiter that manages throttling of requests.
        model (str, optional): The model to use. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): The temperature to use. Defaults to 0.
        max_tokens (int, optional): The maximum tokens to use. Defaults to None.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.

    Returns:
        Any: The response from the chat completion.
    """
    async with limiter:
        response = await openai.ChatCompletion.acreate(
            messages=chat_pair.prompt_raw,
            **kwargs
        )
        return chat_pair, response


async def async_create_chat_completion(
    messages: ChatConversation,
    model: str = "gpt-3.5-turbo",
    temperature: float = 0,
    max_tokens: Union[int, None] = None,
    print_every: int = False,
    keep_order: bool = False,
    requests_per_minute: int = 300,
    top_p: Optional[float] = 1,
    n: Optional[int] = 1,
    stop: Optional[str] = None,
    presence_penalty: Optional[float] = 0,
    frequency_penalty: Optional[float] = 0,
) -> ChatConversation:
    """Generate from OpenAI Chat Completion API asynchronously.

    Args:
        messages (ChatConversation): List of full prompts to generate from.
        model (str, optional): Model configuration. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): Temperature to use. Defaults to 0.
        max_tokens (Union[int, None], optional): Maximum number of tokens to generate. Defaults to None.
        print_every (int, optional): Print the response after every this many calls. It does nothing if keep_order is True. Defaults to False.
        keep_order (bool, optional): If True, keeps the order of responses same as the input prompts. Defaults to False.
        requests_per_minute (int, optional): Number of requests per minute to allow. Defaults to 300.
        response_list (List[str], optional): If provided, responses are added to this list.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.

    Returns:
        ChatConversation: List of generated responses and previous responses.
    """
    if keep_order and print_every:
        print("print_every will do nothing since keep_order is True")
    api_manager = SingleApiManager()
    openai.aiosession.set(ClientSession())
    limiter = aiolimiter.AsyncLimiter(requests_per_minute)
    
    kwargs = {
        "model": model,
        "temperature": temperature,
        "max_tokens": max_tokens,
        "top_p": top_p,
        "n": n,
        "stop": stop,
        "presence_penalty": presence_penalty,
        "frequency_penalty": frequency_penalty,
    }
    
    async_responses = [
        _throttled_acreate_chat_completion(
            **kwargs,
            limiter=limiter,
            chat_pair=chat_pair,
        )
        for chat_pair in messages.conversation_history
    ]

    responses = []
    prompt_tokens = 0
    completion_tokens = 0

    if keep_order:
        # Keep the order of responses same as input prompts
        collect_responses = await tqdm_asyncio.gather(*async_responses)

        # Update the cost associated with this response in the API manager
        for i, response in enumerate(collect_responses):
            chat_pair, openai_response = response
            responses.append(chat_pair.update_response(ResponseDict(openai_response)))
            prompt_tokens += openai_response.usage.prompt_tokens
            completion_tokens += openai_response.usage.completion_tokens

            if i == len(collect_responses) - 1:
                api_manager.update_cost(
                    prompt_tokens,
                    completion_tokens,
                    openai_response.model,
                )
    else:
        # Without order
        for future in tqdm_asyncio.as_completed(async_responses):
            collect_response = await future
            
            chat_pair, openai_response = collect_response

            # Update the cost associated with this response in the API manager
            api_manager.update_cost(
                openai_response.usage.prompt_tokens,
                openai_response.usage.completion_tokens,
                openai_response.model,
            )

            response_text = openai_response.choices[0].message.content
            responses.append(chat_pair.update_response(ResponseDict(openai_response)))
            if print_every:
                print(response_text + "\n")

    # Close the session
    await openai.aiosession.get().close()  # type: ignore
    return messages.fill_conversation(responses)



import argparse
import nbformat as nbf
from typing import Optional
from aitemplates.cells import get_cell1_content, get_cell2_content, get_cell3_content, get_cell4_content, get_cell5_content


def create_notebook(
    filename: str, db: Optional[bool] = False, asnc: Optional[bool] = False, func: Optional[bool] = False
):
    """Creates a notebook using nbformat and populate the first 3 cells

    Args:
        filename (str): The name of the notebook
        db (bool, optional): Whether to use a vector database. Defaults to False.
        asnc (bool, optional): Whether to use async chat completion. Defaults to False.
        func (bool, optional): Whether to use function completion. Defaults to False.

    Returns:
        None
    """
    

    nb = nbf.v4.new_notebook()

    cell1_content = get_cell1_content(db, asnc, func)
    
    cell1 = nbf.v4.new_code_cell(cell1_content)
    
    cell2_content = get_cell2_content(db, asnc, func)
    
    cell2 = nbf.v4.new_code_cell(cell2_content)
    
    cell3_content = get_cell3_content(db, asnc, func)
    
    cell3 = nbf.v4.new_code_cell(cell3_content)
    
    cell4_content = get_cell4_content(db, asnc, func)
    
    cell4 = nbf.v4.new_code_cell(cell4_content)
    
    cell5_content = get_cell5_content(db, asnc, func)
    
    cell5 = nbf.v4.new_code_cell(cell5_content)

    # Add cells to notebook
    nb.cells = [cell1, cell2, cell3, cell4, cell5]

    with open(filename, "w") as f:
        nbf.write(nb, f)


def main():
    parser = argparse.ArgumentParser(description="Generate a new Jupyter notebook")
    parser.add_argument("filename", help="Name of the notebook to create")
    parser.add_argument(
        "-db", action="store_true", help="With or without a vector database"
    )
    parser.add_argument(
        "-asnc",
        action="store_true",
        help="Async chat completion instead of a regular one",
    )
    parser.add_argument(
        "-func",
        action="store_true",
        help="Include openai function templates in the notebook",
    )
    args = parser.parse_args()
    
    if args.asnc and args.func:
        print("openai async method create does not have a function parameter")
        return

    create_notebook(f"{args.filename}.ipynb", args.db, args.asnc, args.func)


if __name__ == "__main__":
    main()



from .oai.responses.chat_response import create_chat_completion
from .oai.responses.async_chat_response import async_create_chat_completion
from .oai.responses.embedding import get_embedding

from .oai.ApiManager import ApiManager, SingleApiManager

from .oai.types.base import Message, ResponseDict
from .oai.types.chat import ChatSequence, ChatConversation, ChatPair
from .oai.types.functions import FunctionPair, FunctionsAvailable


def get_cell1_content(db, asnc, func):
    db1_content = """
chroma_client = chromadb.Client()

embedder = embedding_functions.OpenAIEmbeddingFunction(
    api_key=OPENAI_API_KEY,
    model_name="text-embedding-ada-002",
)

collection = chroma_client.create_collection(name="", embedding_function=embedder)"""

    imports = "from aitemplates import Message, ChatSequence"
    
    if asnc or func:
        imports += ", ChatConversation"
        
    if asnc:
        imports += ", async_create_chat_completion"
    else:
        imports += ", create_chat_completion"
    if func:
        imports += ", FunctionPair, FunctionsAvailable"
        
    if db:
        imports += """
import os
import chromadb
from chromadb.utils import embedding_functions"""
        
    
    cell1_content = f"""\
{imports}
import openai
import os
from dotenv import load_dotenv

load_dotenv()

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

openai.api_key = OPENAI_API_KEY

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")
{db1_content if db else ""}"""
    return cell1_content


def get_cell2_content(db, asnc, func):
    additional = ""
    
    if asnc or func:
        additional+= """
sequence2 = ChatSequence([user_query_msg])
chat = ChatConversation([sequence1, sequence2])"""
        
    cell2_content = f"""\
system_prompt=""
system_prompt_msg = Message("system", system_prompt)

description = ""
description_msg = Message("system", description)

user_query=""
user_query_msg = Message("user", user_query)

sequence1 = ChatSequence([system_prompt_msg, description_msg, user_query_msg])
{additional}"""
    return cell2_content


def get_cell3_content(db, asnc, func):
    additional = ""
    
    if func:
        additional+="""\
func_desc = {
        "name": "func_name",
        "description": "",
        "parameters": {
            "type": "object",
            "properties": {
                "property1": {
                    "type": "string",
                    "description": "",
                },
            },
            "required": ["property1"]
        },
    }
    
def func(property1: str) -> float:
    return 0.0
    
# match the description to the function
function_pair1 = FunctionPair(func_desc, func)

# add FunctionsAvailable dataclass for easy access
functions_available = FunctionsAvailable([function_pair1])
"""
        
    if asnc:
        additional+=f"\nasync_response = await async_create_chat_completion(chat, keep_order=True)"
    else:
        additional+=f"\ncompletion = create_chat_completion(sequence1{', functions=functions_available' if func else ''})"

    cell3_content = f"""\
{additional}"""
    return cell3_content

def get_cell4_content(db, asnc, func):
    additional = ""
    
    if asnc and db:
        additional+="""\
collection.add(
    documents=[*async_response.get_last_responses()], # takes param of num of responses, or call .get_last_responses(all=True)
    metadatas=[{"": ""}, {"": ""}],
    ids=["", ""]
)"""
    elif db:
        additional += """
collection.add(
    documents=[completion],
    metadatas=[{"": ""}],
    ids=[""]
)"""

    cell4_content = f"""\
{additional}
{"async_response.display_conversation()" if asnc else "completion"}"""
    return cell4_content

def get_cell5_content(db, asnc, func):
    cell5_content = f"""\
from aitemplates import SingleApiManager

api_manager = SingleApiManager()
api_manager.total_cost"""
    return cell5_content